// .po file like language pack
plupload.addI18n({
	'Select files' : 'Vyberte súbory',
	'Add files to the upload queue and click the start button.' : 'Pridajte súbory do zoznamu a potom spustite nahrávanie.',
	'Filename' : 'Názov súboru',
	'Status' : 'Stav',
	'Size' : 'Veľkosť',
	'Add files' : 'Pridať súbory',
	'Stop current upload' : 'Zastaviť nahrávanie',
	'Start uploading queue' : 'Spustiť nahrávanie zoznamu',
	'Drag files here.' : 'Sem pretiahnite súbory.',
	'Start upload': 'Spustiť nahrávanie',
	'Uploaded %d/%d files': 'Nahraných %d/%d súborov',
	'Using runtime: ': 'K odoslaniu súborov sa použije rozhranie: ',
	'N/A' : 'N/A',
	'File extension error.': 'Chybný typ súboru.',
	'File size error.': 'Súbor je príliš veľký.',
	'Init error.': 'Chyba inicializácie.',
	'HTTP Error.': 'HTTP Chyba.',
	'Security error.': 'Bezpečnostná Chyba.',
	'Generic error.': 'Chyba.',
	'IO error.': 'IO Chyba',
	'Stop Upload': 'Zastaviť nahrávanie',
	'%d files queued': '%d súborov pridaných do zoznamu'
});