<Application
	xmlns="http://schemas.microsoft.com/winfx/2006/xaml/presentation"
	xmlns:x="http://schemas.microsoft.com/winfx/2006/xaml"
	x:Class="Moxiecode.Plupload.App">
	<Application.Resources>
		<!-- Resources scoped at the Application level should be defined here. -->
	</Application.Resources>
</Application>
